import katapult
import asyncio
import sys 
import psutil
import subprocess
import multiprocessing 
import time
import copy
from katapult.maestroserver import main as server_main
from katapult.maestroclient import maestro_client
from katapult.provider import make_client_command , stream_dump  

# if [[ $(ps aux | grep "katapult.maestroserver" | grep -v 'grep') ]] ; then
#     echo "Katapult server already running"
# else
#     if [[ $(ps -ef | awk '/[m]aestroserver/{print $2}') ]] ; then
#         ps -ef | awk '/[m]aestroserver/{print $2}' | xargs kill 
#     fi
#     echo "Starting Katapult server ..."
#     python3 -u -m katapult.maestroserver
# fi

def katapult_kill(p):
    try:
        p.kill()
    except:
        pass
    try:
        p.terminate()
    except:
        pass

def is_katapult_process(p,name='maestroserver'):
    try:
        if name in p.name():
            return True
        for arg in p.cmdline():
            if name in arg:
                return True
        return False
    except psutil.AccessDenied:
        return False
    except psutil.NoSuchProcess:
        return False

def cli(command):
    start_server() # locally we use python to start the server. maximizes the chance to be Windows complatible
    maestro_client(command)

def start_server():
    server_started = False
    for pid in psutil.pids():
        try:
            p = psutil.Process(pid)
            if is_katapult_process(p,"katapult.maestroserver"):
                print("[Katapult server already started]")
                server_started = True
                break
        except psutil.NoSuchProcess:
            pass
    if not server_started:
        # make sure we kill any rogue processes
        for pid in psutil.pids():
            try:
                p = psutil.Process(pid)
                for pp in p.children(recursive=True):
                    if is_katapult_process(pp):
                        katapult_kill(pp)
                if is_katapult_process(p):
                    katapult_kill(p)
            except psutil.NoSuchProcess:
                pass
        print("[Starting Katapult server ...]")
        subprocess.Popen(['python3','-u','-m','katapult.maestroserver'])
        time.sleep(1)
        #q = multiprocessing.Queue()
        #p = multiprocessing.Process(target=server_main,args=(q,))
        #p = multiprocessing.Process(target=server_main)
        #p.daemon = True
        #p.start()

def cli_translate(command,args):
    if command == 'init':
        return args

    elif command == 'wakeup':
        return args

    elif command == 'start':
        if not args:
            return [False]
        elif len(args)==1:
            return [args[0].strip().lower() == "true"]
        else:
            return [False]

    elif command == 'cfg_add_instances':
        return args

    elif command == 'cfg_add_environments':
        return args

    elif command == 'cfg_add_jobs':
        return args

    elif command == 'cfg_add_config':
        return args

    elif command == 'cfg_reset':
        return args

    elif command == 'deploy':
        return args

    elif command == 'run':
        return args

    elif command == 'kill':
        return args
    
    elif command == 'wait':
        new_args = []
        if args and len(args) >= 1:
            job_state = int(args[0])
            new_args.append(job_state)
        if args and len(args) >= 2:
            run_session = KatapultRunSessionProxy( args[1] )
            new_args.append( stream_dump(run_session) )
        return new_args 

    elif command == 'get_num_active_processes':

        new_args = []
        if args and len(args) >= 1:
            run_session = KatapultRunSessionProxy( args[0] )
            new_args.append( stream_dump(run_session) )
        return new_args 

    elif command == 'get_num_instances':
        return args

    elif command == 'get_states' or command == 'get_jobs_states':
        new_args = []
        if args and len(args) >= 1:
            run_session = KatapultRunSessionProxy( args[0] )
            new_args.append( stream_dump(run_session) )
        return new_args 

    elif command == 'print_summary' or command == 'print':
        new_args = []
        if args and len(args) >= 1:
            run_session = KatapultRunSessionProxy( args[0] )
            new_args.append( stream_dump(run_session) )
        if args and len(args) >= 2:
            instance = KatapultInstanceProxy( args[1] )
            new_args.append( stream_dump(instance) )
        return new_args 

    elif command == 'print_aborted' or command == 'print_aborted_logs':
        new_args = []
        if args and len(args) >= 1:
            run_session = KatapultRunSessionProxy( args[0] )
            new_args.append( stream_dump(run_session) )
        if args and len(args) >= 2:
            instance = KatapultInstanceProxy( args[1] )
            new_args.append( stream_dump(instance) )
        return new_args 
        
    elif command == 'print_objects':
        return args

    elif command == 'clear_results_dir':
        return args

    # out_dir=None,run_session=None,use_cached=True,use_normal_output=False
    elif command == 'fetch_results':

        new_args = []
        if args and len(args)>=1:
            directory = args[0].strip()
            if directory.lower() == 'none':
                directory = None
            new_args.append(directory)
        if args and len(args)>=2:
            run_session = KatapultRunSessionProxy( args[1] )
            new_args.append( stream_dump(run_session) )
        if args and len(args)>=3:
            use_cached = args[2].lower().strip() == "true"
            new_args.append( use_cached )
        if args and len(args)>=4:
            use_normal_output = args[3].lower().strip() == "true"
            new_args.append( use_normal_output )
        return new_args

    elif command == 'finalize':
        return args

    elif command == 'shutdown':
        return args

    elif command == 'test':
        return args
    
    else:
        return None

def main():
    multiprocessing.set_start_method('spawn')

    if len(sys.argv)<2:
        print("python3 -m katapult.cli CMD [ARGS]")
        sys.exit()

    args = copy.deepcopy(sys.argv)
    cmd_arg = args.pop(0) #trash
    while 'cli' not in cmd_arg:
        cmd_arg = args.pop(0)
    command = args.pop(0)

    ser_args = cli_translate(command,args)
    # lets not escape the command, we're not sending it to a stream
    the_command = make_client_command( command , ser_args , False)
    cli(the_command)


if __name__ == '__main__':     
    main()



from abc import ABC , abstractmethod
from katapult.provider import KatapultProvider , stream_load , stream_dump , line_buffered , make_client_command , STREAM_RESULT, DIRECTORY_TMP , get_EOL_conversion
from katapult.core import *
from katapult.attrs import *
import copy , io
from zipfile import ZipFile
import sys , os , fnmatch
import re , json
from os.path import basename
import pkg_resources
import time
import random
import shutil
import asyncssh
import asyncio
import platform

random.seed()

####################################
# Client handling MAESTRO instance #
####################################

class KatapultLightProvider(KatapultProvider,ABC):

    def __init__(self,conf=None):

        KatapultProvider.__init__(self,conf)

        #self._install_maestro()

    def _init(self,conf):
        self.ssh_conn   = None
        self.ftp_client = None
        self._current_session = None
        super()._init(conf)
        self._load()
    
    def _load(self):
        self._maestro = None
        if self._config.get('maestro')=='remote':
            if not self._config.get('instances') or len(self._config.get('instances'))==0:
                self.debug(2,"There are no instances to watch - skipping maestro creation")
                return
            else:
                # let's try to match the region of the first instance ...
                region = None
                if len(self._config.get('instances')) > 0:
                    region = self._config.get('instances')[0].get('region')
                if not region:
                    region = self.get_region()

                img_id , img_user , img_type = self.get_suggested_image(region)
                if not img_id:
                    self.debug(1,"Using first instance information to create MAESTRO")
                    img_id   = self._config.get('instances')[0].get('img_id')
                    img_user = self._config.get('instances')[0].get('img_username')
                    img_type = self._config.get('instances')[0].get('img_type')
                
                maestro_cfg = { 
                    'maestro'      : True ,
                    'img_id'       : img_id ,
                    'img_username' : img_user ,                 
                    'type'         : img_type , 
                    'dev'          : self._config.get('dev',False) ,
                    'project'      : self._config.get('project',None) ,
                    'region'       : region ,
                    '_maestro_name_proj' : self._config.get('_maestro_name_proj',False)
                }
                self._maestro = KatapultInstance(maestro_cfg,None)

    async def _deploy_maestro(self,reset):
        # deploy the maestro ...
        if self._maestro is None:
            self.debug(2,"no MAESTRO object to deploy")
            return

        # wait for maestro to be ready
        instanceid , ssh_conn , ftp_client = await self._wait_and_connect(self._maestro)

        home_dir = self._maestro.get_home_dir()
        katapult_dir = self._maestro.path_join( home_dir , 'katapult' )
        files_dir = self._maestro.path_join( katapult_dir , 'files' )
        ready_file = self._maestro.path_join( katapult_dir , 'ready' )
        maestro_file = self._maestro.path_join( home_dir , 'maestro' )
        aws_dir = self._maestro.path_join( home_dir , '.aws' )
        aws_config = self._maestro.path_join( aws_dir , 'config' )
        if self._maestro.get_platform() == KatapultPlatform.LINUX or self._maestro.get_platform() == KatapultPlatform.WINDOWS_WSL :
            activate_file = self._maestro.path_join( katapult_dir , '.venv' , 'maestro' , 'bin' , 'activate' )
        elif self._maestro.get_platform() == KatapultPlatform.WINDOWS:
            activate_file = self._maestro.path_join( katapult_dir , '.venv' , 'maestro' , 'Scripts' , 'activate.bat' )
        elif self._maestro.get_platform() == KatapultPlatform.UNKNOWN:
            if 'windows' in platform.system().lower():
                activate_file = os.path.join( katapult_dir , '.venv' , 'maestro' , 'Scripts' , 'activate.bat' )
            else:
                activate_file = self._maestro.path_join( katapult_dir , '.venv' , 'maestro' , 'bin' , 'activate' )
        elif self._maestro.get_platform() == KatapultPlatform.MOCK:
            if 'windows' in platform.system().lower():
                activate_file = os.path.join( katapult_dir , '.venv' , 'maestro' , 'Scripts' , 'activate.bat' )
            else:
                activate_file = self._maestro.path_join( katapult_dir , '.venv' , 'maestro' , 'bin' , 'activate' )

        re_init  = await self._test_reupload(self._maestro,ready_file, ssh_conn)

        if re_init:
            # remove the file
            stdout , stderr , ssh_conn , ftp_client = await self._exec_maestro_command_simple(ssh_conn,ftp_client,'rm -f ' + ready_file )
            # make katapult dir
            stdout , stderr , ssh_conn , ftp_client = await self._exec_maestro_command_simple(ssh_conn,ftp_client,'mkdir -p ' + files_dir ) 
            # mark it as maestro...
            stdout , stderr , ssh_conn , ftp_client = await self._exec_maestro_command_simple(ssh_conn,ftp_client,'echo "" > ' + maestro_file ) 
            # add manually the 
            if self._config.get('profile'):
                profile = self._config.get('profile')
                region  = self.get_region() # we have a profile so this returns the region for this profile
                aws_config_cmd = "mkdir -p "+aws_dir+" && echo \"[profile "+profile+"]\nregion = " + region + "\noutput = json\" > " + aws_config
                stdout , stderr , ssh_conn , ftp_client =  await self._exec_maestro_command_simple(ssh_conn,ftp_client,aws_config_cmd)
            # grant its admin rights (we need to be (stopped or) running to be able to do that)
            self.grant_admin_rights(self._maestro)
            # setup auto_stop behavior for maestro
            self.setup_auto_stop(self._maestro)
            # deploy Katapult on the maestro
            await self._deploy_katapult(ssh_conn,ftp_client)
            # mark as ready
            stdout , stderr , ssh_conn , ftp_client = await self._exec_maestro_command_simple(ssh_conn,ftp_client,'if [ -f '+activate_file+' ]; then echo "" > '+ready_file+' ; fi')

        # deploy the config to the maestro (every time)
        await self._deploy_config(ssh_conn,ftp_client)

        # let's redeploy the code every time for now ... (if not already done in re_init)
        if not re_init and reset:
            await self._deploy_katapult_files(ssh_conn,ftp_client)

        # start the server (if not already started)
        await self._run_server(ssh_conn)

        # wait for maestro to have started
        await self._wait_for_maestro(ssh_conn)

        #time.sleep(30)

        self.ssh_conn = ssh_conn
        self.ftp_client = ftp_client
        #ftp_client.close()
        #ssh_conn.close()

        self.debug(1,"MAESTRO is READY",color=bcolors.OKCYAN)


    async def _deploy_katapult_files(self,ssh_conn,ftp_client):
        await ftp_client.chdir(self._get_katapult_dir())
        # CP437 is IBM zip file encoding
        await self.sftp_put_bytes(ftp_client,'katapult.zip',self._create_katapult_zip(),'CP437')
        commands = [
            { 'cmd' : 'cd '+self._get_katapult_dir()+' && unzip -o katapult.zip && rm katapult.zip' , 'out' : True } ,
        ]
        await self._run_ssh_commands(self._maestro,ssh_conn,commands) 
        
        filesOfDirectory = os.listdir('.')
        pattern = "katapult*.pem"
        for file in filesOfDirectory:
            if fnmatch.fnmatch(file, pattern):
                await ftp_client.put(os.path.abspath(file),os.path.basename(file))
        sh_files = self._get_remote_files_path( '*.sh')
        commands = [
            { 'cmd' : 'chmod +x ' + sh_files , 'out' : True } ,
        ]
        eol_command = get_EOL_conversion(self._maestro,sh_files)
        if eol_command:
            commands.append({'cmd':eol_command,'out':True})

        await self._run_ssh_commands(self._maestro,ssh_conn,commands) 


    async def _deploy_katapult(self,ssh_conn,ftp_client):

        # upload katapult files
        await self._deploy_katapult_files(ssh_conn,ftp_client)
        
        maestroenv_sh = self._get_remote_files_path( 'maestroenv.sh' )
        # unzip katapult files, install and run
        commands = [
            { 'cmd' : maestroenv_sh , 'out' : True } ,
        ]
        await self._run_ssh_commands(self._maestro,ssh_conn,commands)

    async def _run_server(self,ssh_conn):
        # run the server
        startmaestro_sh = self._get_remote_files_path( 'startmaestro.sh' )
        commands = [
            { 'cmd' : startmaestro_sh , 'out' : False , 'output' : 'maestro.log' },
            { 'cmd' : 'crontab -r ; echo "* * * * * '+startmaestro_sh+' auto_init" | crontab', 'out' : True }
        ]
        await self._run_ssh_commands(self._maestro,ssh_conn,commands)

        # we have separated run and init and start 
        # we need to init with the deployed config
        self.debug(1,"initialize server with config")
        init_objects = await self._exec_maestro_command("init","config.json")
        
        self._update_K_loaded(init_objects)

        # TODO: improve return result from init and match whats been loaded and what has not
        # (same as cfg_add_***)
        # for objs_key in ['instances','environments','jobs']:
        #     if self._config.get(objs_key):
        #         for obj_cfg in self._config[objs_key]:
        #             obj_cfg[K_LOADED]=True

    async def _deploy_config(self,ssh_conn,ftp_client,config_filename='config.json',only_new=False):

        config , mkdir_cmd , files_to_upload_per_dir = self._translate_config_for_maestro(only_new)

        # we can add some keyed arguments to the config ...
        # useful to serialize some arguments between light client and fat client
        # if kwargs and len(kwargs)>0:
        #     config['_kwargs'] = dict()
        #     for k,v in kwargs.items():
        #         config['_kwargs'][k] = stream_dump(v)

        # serialize the config and send it to the maestro
        await ftp_client.chdir(self._get_katapult_dir())
        await self.sftp_put_string(ftp_client,config_filename,json.dumps(config))
        # execute the mkdir_cmd
        if mkdir_cmd:
            stdout , stderr , ssh_conn , ftp_client = await self._exec_maestro_command_simple(ssh_conn,ftp_client,mkdir_cmd)
            await stdout.read()
        for remote_dir , files_infos in files_to_upload_per_dir.items():
            await ftp_client.chdir(remote_dir)
            for file_info in files_infos:
                remote_file = os.path.basename(file_info['remote'])
                try:
                    await ftp_client.put(file_info['local'],remote_file)
                except FileNotFoundError as fnfe:
                    self.debug(1,"You have specified a file that does not exist:",fnfe,color=bcolors.FAIL)
                except asyncssh.sftp.SFTPNoSuchFile as nsf:
                    self.debug(1,"You have specified a file that does not exist:",nsf,remote_dir,file_info,color=bcolors.FAIL)
        return config 

    def _translate_config_for_maestro(self,only_new):
        if only_new:
            config = dict()
            for master_key in self._config:
                if master_key in K_OBJECTS:
                    config[master_key] = []                    
                    for obj_cfg in self._config[master_key]:
                        if obj_cfg.get(K_LOADED) == False:
                            config[master_key].append(copy.deepcopy(obj_cfg))
                else:
                    config[master_key] = copy.deepcopy(self._config[master_key])
        else:
            config = copy.deepcopy(self._config)

        config['maestro'] = 'local' # the config is for maestro, which needs to run local

        config['region'] = self.get_region()

        if config.get('instances'):
            for i , inst_cfg in enumerate(config.get('instances')):
                # we need to freeze the region within each instance config ...
                if not config['instances'][i].get('region'):
                    config['instances'][i]['region'] = self.get_region()

        if config.get('environments'):
            for i , env_cfg in enumerate(config.get('environments')):
                # Note: we are using a simple KatapultEnvironment here 
                # we're not deploying at this stage
                # but this will help use re-use all the business logic in katapultutils envs
                # in order to standardize the environment and create an inline version 
                # through the KatapultEnvironment::json() method
                env = KatapultEnvironment(self._config.get('project'),env_cfg)
                config['environments'][i] = env.get_env_obj()
                config['environments'][i][K_COMPUTED] = True

        files_to_upload = dict()

        if config.get('jobs'):
            for i , job_cfg in enumerate(config.get('jobs')):
                if job_cfg.get('run_script'):
                    run_script = job_cfg.get('run_script')  
                    ref_file0  = run_script
                    args       = ref_file0.split(' ')
                    if args and len(args)>0:
                        ref_file0 = args[0]
                else:
                    ref_file0 = None
                #for attr,use_ref in [ ('run_script',False) , ('upload_files',True) , ('input_files',True) , ('output_files',True) ] :
                for attr,use_ref in [ ('run_script',False) , ('upload_files',True) , ('input_files',True) ] :
                    upfiles = job_cfg.get(attr)
                    ref_file = ref_file0 if use_ref else None
                    if not upfiles:
                        continue
                    if isinstance(upfiles,str):
                        upfiles_split = upfiles.split(' ')
                        if len(upfiles_split)>0:
                            upfiles = upfiles_split[0]
                    if isinstance(upfiles,str):
                        local_abs_path , local_rel_path , remote_abs_path , remote_rel_path , external = self._resolve_maestro_job_paths(upfiles,ref_file,self._get_home_dir())
                        if local_abs_path not in files_to_upload:
                            files_to_upload[local_abs_path] = { 'local' : local_abs_path , 'remote' : remote_abs_path }
                        if attr == 'run_script':
                            args = run_script.split(' ')
                            args[0] = remote_abs_path
                            config['jobs'][i][attr] = (' ').join(args)
                        else:
                            config['jobs'][i][attr] = remote_abs_path
                    else:
                        config['jobs'][i][attr] = []
                        for upfile in upfiles:
                            local_abs_path , local_rel_path , remote_abs_path , remote_rel_path , external = self._resolve_maestro_job_paths(upfile,ref_file,self._get_home_dir())
                            if local_abs_path not in files_to_upload:
                                files_to_upload[local_abs_path] = { 'local' : local_abs_path , 'remote' : remote_abs_path }
                            config['jobs'][i][attr].append(remote_abs_path)
        
        mkdir_cmd = ""
        files_to_upload_per_dir = dict()
        for key,file_info in files_to_upload.items():
            remote_dir = os.path.dirname(file_info['remote'])
            if remote_dir not in files_to_upload_per_dir:
                files_to_upload_per_dir[remote_dir] = []
                mkdir_cmd += "mkdir -p " + remote_dir + ";"
            files_to_upload_per_dir[remote_dir].append(file_info)

        self.debug(2,config)

        return config , mkdir_cmd , files_to_upload_per_dir

    def _resolve_maestro_job_paths(self,upfile,ref_file,home_dir):
        files_path = self._maestro.path_join( home_dir , 'files' )
        return katapultutils.resolve_paths(self._maestro,upfile,ref_file,files_path,True) # True for mutualized 

    def _get_home_dir(self):
        return self._maestro.get_home_dir()

    def _get_katapult_dir(self):
        return self._maestro.path_join( self._get_home_dir() , 'katapult' )

    def _get_remote_files_path(self,files=None):
        if not files:
            return self._maestro.path_join( self._get_katapult_dir() , 'katapult' , 'resources' , 'remote_files' )        
        else:
            return self._maestro.path_join( self._get_katapult_dir() , 'katapult' , 'resources' , 'remote_files' , files )        

    def _zip_package(self,package_name,src,dest,zipObj):
        dest = os.path.join( dest , os.path.basename(src) ).rstrip( os.sep )
        if pkg_resources.resource_isdir(package_name, src):
            #if not os.path.isdir(dest):
            #    os.makedirs(dest)
            for res in pkg_resources.resource_listdir(package_name, src):
                self.debug(2,'scanning package resource',res)
                self._zip_package(package_name,os.path.join( src , res), dest, zipObj)
        else:
            if os.path.splitext(src)[1] not in [".pyc"] and not src.strip().endswith(".DS_Store"):
                #copy_resource_file(src, dest) 
                data_str = pkg_resources.resource_string(package_name, src)
                self.debug(2,"Writing",src)
                zipObj.writestr(dest,data_str)

    def _create_katapult_zip(self):
        zip_buffer = io.BytesIO()
        # create a ZipFile object
        with ZipFile(zip_buffer, 'w') as zipObj:    
            import katapult as kt
            katapultinit = os.path.abspath(kt.__file__) # this is the __init__.py file
            katapultmodu = os.path.dirname(katapultinit)
            katapultroot = os.path.dirname(katapultmodu)
            for otherfilepath in [ 'requirements.txt' ]:
                opened = False
                for thedir in [katapultmodu , katapultroot ]: # not sure why: Kyel bug Windows
                    thefilepath = os.path.join(thedir,otherfilepath)
                    if os.path.exists(thefilepath):
                        try:
                            with open(thefilepath,'r') as thefile:
                                opened = True
                                data = thefile.read()
                                zipObj.writestr(otherfilepath,data)
                        except:
                            self.debug(1,"Could not open file ",thefilepath,color=bcolors.FAIL)
                            pass
                if not opened:
                    self.debug(1,"Could not open file ",otherfilepath,color=bcolors.FAIL)
            # write the package
            self._zip_package("katapult",".","katapult",zipObj)

        self.debug(2,"ZIP BUFFER SIZE =",zip_buffer.getbuffer().nbytes)

        # very important...
        zip_buffer.seek(0)

        return zip_buffer.getvalue() #.getbuffer()

    async def _wait_for_maestro(self,ssh_conn):
        #if self.ssh_conn is None:
        #    instanceid , self.ssh_conn , self.ftp_client = self._wait_and_connect(self._maestro)
        
        waitmaestro_sh = self._get_remote_files_path( 'waitmaestro.sh' )
        cmd = waitmaestro_sh+" 1" # 1 = with tail log

        stdout , stderr , ssh_conn , ftp_client = await self._exec_maestro_command_simple(ssh_conn,None,cmd)      

        try:
            async for line in stdout:
                if not line:
                    break
                self.debug(1,line,end='')
        except asyncssh.misc.ConnectionLost:
            pass
        except StopAsyncIteration:
            pass

        # for l in await line_buffered(stdout):
        #     if not l:
        #         break
        #     self.debug(1,l,end='')

    async def _exec_maestro_command_simple(self,ssh_conn,ftp_client,command_line):
        if ssh_conn is None:
            instanceid , ssh_conn , ftp_client = await self._wait_and_connect(self._maestro)
        retrys = 0 
        while True:
            try:
                stdout,stderr = await self._exec_command(ssh_conn,command_line)      
                return stdout,stderr,ssh_conn,ftp_client
            except (OSError, asyncssh.Error) as e:
                retrys += 1
                if retrys<5:
                    self.debug(1,"Retrying ...",color=bcolors.WARNING)
                    self.debug(2,"Retrying command",maestro_command,color=bcolors.WARNING)
                    await asyncio.sleep(10)
                    instanceid , ssh_conn , ftp_client = await self._wait_and_connect(self._maestro)
                    await self._run_server(ssh_conn)
                else:
                    self.debug(1,"Enough Retries",e,color=bcolors.FAIL)
                    break     
        return None,None,ssh_conn,ftp_client 

    async def _exec_maestro_command(self,maestro_command,raw_args=None):

        # args = None
        # if maestro_command != "init":
        #     if raw_args:
        #         args = []
        #         if isinstance(raw_args,list):
        #             for arg in raw_args:
        #                 # we dont dump the array cause we want to use the separators __,__
        #                 # so we dump the individual array elements
        #                 args.append( stream_dump(arg) ) 
        #         else:
        #             args = stream_dump(raw_args)
        # else:
        #     args = raw_args

        args = stream_dump(raw_args)

        the_command = make_client_command(maestro_command,args)

        if self.ssh_conn is None:
            instanceid , self.ssh_conn , self.ftp_client = await self._wait_and_connect(self._maestro)
        
        # -u for no buffering
        waitmaestro_sh = self._get_remote_files_path( 'waitmaestro.sh' )
        venv_python    = self._maestro.path_join( self._get_katapult_dir() , '.venv' , 'maestro' , 'bin' , 'python3' )

        cmd = "cd "+self._get_katapult_dir()+ " && "+waitmaestro_sh+" && sudo "+venv_python+" -u -m katapult.maestroclient " + the_command

        retrys = 0 

        while True:
            try:
                stdout , stderr = await self._exec_command(self.ssh_conn,cmd)    
                try:
                    async for line in stdout:
                        if not line:
                            break
                        # kinda dirty ... 
                        if line.startswith(STREAM_RESULT):
                            line = line.replace(STREAM_RESULT,'')
                            line = line.strip()
                            self.debug(2,"Got RESULT",line,color=bcolors.OKCYAN)
                            return stream_load(self,json.loads(line)) # result is printed last...
                        else:
                            self.debug(1,line,end='')
                except asyncssh.misc.ConnectionLost:
                    pass
                except StopAsyncIteration:
                    pass
                except:
                    pass
                
                break
            except (OSError, asyncssh.Error):
                retrys += 1
                if retrys<5:
                    await asyncio.sleep(10)
                    self.debug(1,"Retrying _exec_maestro_command",cmd,color=bcolors.WARNING)
                    instanceid , self.ssh_conn , self.ftp_client = await self._wait_and_connect(self._maestro)
                    await self._run_server(self.ssh_conn)
                else:
                    self.debug(1,"Enough tries _exec_maestro_command",color=bcolors.FAIL)
                    break


        return None

        # for l in await line_buffered(stdout):
        #     if not l:
        #         break
        #     self.debug(1,l,end='')
        
        # while True:
        #     outlines = stdout.readlines()
        #     if not outlines:
        #         errlines = stderr.readlines()
        #         for eline in errlines:
        #             self.debug(1,eline,end='')
        #         break
        #     for line in outlines:
        #         self.debug(1,line,end='')
        #     errlines = stderr.readlines()
        #     for eline in errlines:
        #         self.debug(1,eline,end='')

        # for l in line_buffered(stdout):
        #     if not l:
        #         errlines = stderr.readlines()
        #         for eline in errlines:
        #             self.debug(1,eline,end='')
        #         break
        #     self.debug(1,l,end='')
        #     errlines = stderr.readlines()
        #     for eline in errlines:
        #         self.debug(1,eline,end='')


    async def _install_maestro(self,reset):
        # this will block for some time the first time...
        self.debug_set_prefix(bcolors.BOLD+'INSTALLING MAESTRO: '+bcolors.ENDC)
        self._instances_states = dict()        
        self._start_and_update_instance(self._maestro)
        if reset:
            await self.reset_instance(self._maestro)
        await self._deploy_maestro(reset) # deploy the maestro now !
        self.debug_set_prefix(None)       

    async def start(self,reset=False):
        # install maestro materials
        await self._install_maestro(reset)
        # triggers maestro::start
        await self._exec_maestro_command("start",reset)

    def _update_K_loaded(self,objs_result):
        for objs_key in K_OBJECTS:
            if objs_result.get(objs_key) and self._config.get(objs_key):
                for obj in objs_result[objs_key]:
                    o_uid = obj.get_config(K_CFG_UID)
                    if not o_uid:
                        continue
                    for cfg in self._config[objs_key]:
                        c_uid = cfg[K_CFG_UID]
                        if o_uid == c_uid:
                            cfg[K_LOADED] = obj.get_config(K_LOADED) 
                            break           

    async def _cfg_add_objects(self,conf,coroutine,name,**kwargs):
        # complement self._config
        await coroutine(conf,**kwargs)
        # wait for maestro to be ready
        instanceid , ssh_conn , ftp_client = await self._wait_and_connect(self._maestro)
        # config_name
        config_name = 'config_add-' + katapultutils.generate_unique_id() + '.json'
        # deploy the new config to the maestro (every time)
        # (this ensures the deployment of dependent files)
        translated_config = await self._deploy_config(ssh_conn,ftp_client,config_name,True) #True = only new stuff
        # triggers maestro::add_* : use string dump or config file name (either way)
        #await self._exec_maestro_command(name,config_name)
        args = [ translated_config , kwargs ]
        objs = await self._exec_maestro_command(name,args)

        # match loaded stuff in the light client config
        self._update_K_loaded(objs)

        return objs

    async def cfg_add_instances(self,conf,**kwargs):
        return await self._cfg_add_objects(conf,super().cfg_add_instances,"cfg_add_instances",**kwargs)

    async def cfg_add_environments(self,conf,**kwargs):
        return await self._cfg_add_objects(conf,super().cfg_add_environments,"cfg_add_environments",**kwargs)

    async def cfg_add_jobs(self,conf,**kwargs):
        return await self._cfg_add_objects(conf,super().cfg_add_jobs,"cfg_add_jobs",**kwargs)

    async def cfg_add_config(self,conf,**kwargs):
        return await self._cfg_add_objects(conf,super().cfg_add_config,"cfg_add_config",**kwargs)

    async def cfg_reset(self):
        # triggers maestro::reset
        await self._exec_maestro_command("cfg_reset") # use output - the deploy part will be skipped depending on option ...

    async def reset_instance(self,instance):
        self.debug(1,'RESETTING instance',instance.get_name())
        instanceid, ssh_conn , ftp_client = await self._wait_and_connect(instance)
        if ssh_conn is not None:
            await self.sftp_put_remote_file(ftp_client,'resetmaestro.sh')
            resetmaestro_sh = self._maestro.path_join( self._maestro.get_home_dir() , 'resetmaestro.sh' )
            commands = []
            eol_command = get_EOL_conversion(instance,resetmaestro_sh)
            if eol_command:
                commands.append({'cmd':eol_command,'out':True})
            commands.append(
               { 'cmd' : 'chmod +x '+resetmaestro_sh+' && ' + resetmaestro_sh , 'out' : True }
            )
            await self._run_ssh_commands(instance,ssh_conn,commands)
            #ftp_client.close()
            ssh_conn.close()
        self.debug(1,'RESETTING done')

    async def deploy(self):
        # triggers maestro::deploy
        await self._exec_maestro_command("deploy") # use output - the deploy part will be skipped depending on option ...

    async def run(self,continue_session=False):
        # triggers maestro::run
        run_session = await self._exec_maestro_command("run",continue_session)
        self._current_session = run_session
        return self._current_session

    async def kill(self,identifier):
        # triggers maestro::kill
        await self._exec_maestro_command("kill",identifier)

    async def wakeup(self):
        await self.start()
        # triggers maestro::wakeup
        await self._exec_maestro_command("wakeup")

    async def wait(self,job_state,run_session=None):
        if run_session is not None:
            args = [ int(job_state) , run_session ]
        else:
            args = [ int(job_state) ]
        # triggers maestro::wait
        await self._exec_maestro_command("wait",args)

    async def get_jobs_states(self,run_session=None,only_ran_processes=False):
        args = [ run_session , only_ran_processes ]
        # triggers maestro::get_jobs_states
        return await self._exec_maestro_command("get_states",args)

    async def print_jobs_summary(self,run_session=None,instance=None):
        if run_session is not None:
            args = [ run_session ]
        else:
            args = []
        if instance is not None:
            args.append(instance.get_name())

        # triggers maestro::print_jobs_summary
        await self._exec_maestro_command("print_summary",args)

    async def print_aborted_logs(self,run_session=None,instance=None):
        if run_session is not None:
            args = [ run_session ]
        else:
            args = []
        if instance is not None:
            args.append(instance.get_name())
        # triggers maestro::print_aborted_logs
        await self._exec_maestro_command("print_aborted",args)

    async def print_objects(self):
        # triggers maestro::print_objects
        await self._exec_maestro_command("print_objects")

    async def clear_results_dir(self,out_dir=None) :

        if out_dir is None:
            out_dir = DIRECTORY_TMP

        if not os.path.isabs(out_dir):
            out_dir = os.path.join(os.getcwd(),out_dir)
        shutil.rmtree(out_dir, ignore_errors=True)    

        if self._maestro:
            homedir     = self._maestro.get_home_dir()
            maestro_dir = self._maestro.path_join( homedir , "katapult_tmp_fetch" )
            await self._exec_maestro_command("clear_results_dir",maestro_dir)


    async def fetch_results(self,out_dir=None,run_session=None,use_cached=True,use_normal_output=False):

        if out_dir is None:
            out_dir = DIRECTORY_TMP

        # out_dir is local
        if not os.path.isabs(out_dir):
            out_dir = os.path.join(os.getcwd(),out_dir)

        if not run_session:
            run_session = self._current_session

        if not run_session:
            self.debug(1,"No session to fetch",color=bcolors.WARNING)
            return None

        session_out_dir  = self._get_session_out_dir(out_dir,run_session)

        # we've already fetched the results (possibly from the watcher process)
        if use_cached and os.path.exists(session_out_dir):
            return session_out_dir

        if session_out_dir and session_out_dir != './':
            try:
                #os.rmdir(out_dir)
                shutil.rmtree(session_out_dir, ignore_errors=True)
            except:
                pass
            try:
                os.makedirs(session_out_dir)
            except:
                pass

        randnum          = str(random.randrange(1000))
        homedir          = self._maestro.get_home_dir()
        maestro_dir      = self._maestro.path_join( homedir , "katapult_tmp_fetch" ) #+ randnum )
        maestro_tar_file = "maestro"+randnum+".tar"
        maestro_tar_path = self._maestro.path_join( homedir , maestro_tar_file )
        session_out_dir_remote = self._get_session_out_dir(maestro_dir,run_session,self._maestro)

        # fetch the results on the maestro
        args = [ maestro_dir , run_session , use_cached , use_normal_output ]
        session_out_dir_remote = await self._exec_maestro_command("fetch_results",args)

        # get the tar file of the results
        stdout , stderr , ssh_conn , ftp_client = await self._exec_maestro_command_simple(None,None,"cd " + session_out_dir_remote + " && tar -cvf "+maestro_tar_path+" .")

        stdout_str = await stdout.read()
        stderr_str = await stderr.read()
        self.debug(1,stdout_str) 
        self.debug(2,stderr_str) 
        local_tar_path = os.path.join(session_out_dir,maestro_tar_file)
        await ftp_client.chdir( homedir )
        await ftp_client.get( maestro_tar_file , local_tar_path )

        # untar
        os.system("tar -xvf "+local_tar_path+" -C "+session_out_dir)

        # cleanup
        os.remove(local_tar_path)
        stdout , stderr , ssh_conn , ftp_client = await self._exec_maestro_command_simple(ssh_conn,ftp_client,"rm -rf "+session_out_dir_remote+" "+maestro_tar_path)

        # close
        #ftp_client.close()
        ssh_conn.close()

        return session_out_dir

    async def finalize(self):
        # triggers maestro::finalize
        try:
            await self._exec_maestro_command("finalize")
        # this will likely happen if auto_stop is one
        except asyncssh.misc.ConnectionLost as cle:
            pass

    async def get_objects(self):
        # triggers maestro::get_objects
        await self._exec_maestro_command("get_objects")

    def _get_or_create_instance(self,instance):
        instance , created = super()._get_or_create_instance(instance)
        # dangerous !
        #self.add_maestro_security_group(instance)
        # open web server

        return instance , created 

    @abstractmethod
    def grant_admin_rights(self,instance):
        pass

    @abstractmethod
    def add_maestro_security_group(self,instance):
        pass

    @abstractmethod
    def setup_auto_stop(self,instance):
        pass

    # needed by KatapultProvider::_wait_for_instance 
    def serialize_state(self):
        pass

    def get_run_session(self,session_id):
        return KatapultRunSessionProxy( session_id )

    def get_instance(self,instance_name,**kwargs):
        return KatapultInstanceProxy( instance_name , kwargs.get('config') )

    def get_environment(self,env_hash,**kwargs):
        return KatapultEnvironmentProxy( env_hash , kwargs.get('config') )

    def get_job(self,job_id,**kwargs):
        return KatapultJobProxy( job_id , kwargs.get('config') )

    async def get_num_active_processes(self,run_session=None):
        if run_session is None:
            run_session = self._current_session
        if not run_session:
            return 0
        args = [ run_session ]
        num_processes = await self._exec_maestro_command("get_num_active_processes",args)
        return int(num_processes)

    async def get_num_instances(self):
        num_instances = await self._exec_maestro_command("get_num_instances")
        return int(num_instances)





from katapult import provider as kt
from katapult.provider import COMMAND_ARGS_SEP , ARGS_SEP , STREAM_RESULT , debug , stream_load , stream_dump
import asyncio , os , sys , time
from katapult.core import KatapultProcessState , bcolors , KatapultRunSession , KatapultRunSessionProxy
import traceback
import json
import socket
import asyncio

HOST = 'localhost' #'0.0.0.0' #127.0.0.1' 
PORT = 5000

async def mainloop(ctxt):

    await ctxt.init()

    server = await asyncio.start_server(ctxt.handle_client, HOST, PORT, reuse_address=True, reuse_port=True)

    addrs = ', '.join(str(sock.getsockname()) for sock in server.sockets)
    print(f'Serving on {addrs}')

    try:
        async with server:
            await server.serve_forever()   
    except asyncio.CancelledError:
        ctxt.restore_stdio()
        print("Shutting down ...")


class ByteStreamWriter():

    def __init__(self,writer):
        self.writer = writer 
    
    def write(self,data):
        if isinstance(data,(bytes,bytearray)):
            self.writer.write(data)
        elif isinstance(data,str):
            self.writer.write(data.encode('utf-8'))
        else:
            self.writer.write(str(data,'utf-8').encode('utf-8'))

    async def drain(self):
        await self.writer.drain()

    def close(self):
        self.writer.close()

    def flush(self):
        asyncio.ensure_future( self.writer.drain() )

class ServerContext:
    def __init__(self,args):
        self.kt_client  = None
        self.old_stdout = sys.stdout
        self.old_stderr = sys.stderr
        self.auto_init = any( arg == 'auto_init' for arg in args )

    async def init(self):
        if self.auto_init: # set to True by the crontab in case maestro crashed
            self.kt_client = kt.get_client(None)
            # we've just restarted a crashed maestro server process
            # let's test for WATCH state and reach it back again if thats needed
            await self.wakeup()

    async def wakeup(self):
        await self.kt_client.wakeup()

    async def handle_client(self, reader, writer):
        while True:
            try:
                cmd_line = await reader.readline()
                if not cmd_line:
                    break
                cmd_line = cmd_line.decode('utf-8').strip()
                #print(cmd_line)

                # OLD SERIALIZATION METHOD (cf. provider.py too)
                # cmd_args = cmd_line.split(COMMAND_ARGS_SEP)
                # if len(cmd_args)>=2:
                #     cmd  = cmd_args[0].strip()
                #     args = cmd_args[1].split(ARGS_SEP)
                # else:
                #     cmd  = cmd_line
                #     args = None

                cmd_json = json.loads(cmd_line)
                # DO NOT UNCOMMENT ! THIS CAUSES A LOCKS
                #print("JSON COMMAND",cmd_json)
                cmd  = cmd_json['cmd']
                args = cmd_json['args'] 

                await self.process_command(cmd,args,writer)
                break # one-shot command
            except ConnectionResetError as cre:
                try:
                    sys.stdout = self.old_stdout
                    print(cre)
                    print("DISCONNECTION")
                except:
                    pass
                break    
            except Exception as e:
                try:
                    sys.stdout = self.old_stdout
                    sys.stderr = self.old_stderr
                    print(e)
                    print(traceback.format_exc())
                    traceback.print_exc()
                except:
                    pass
                break  
        try:
            await writer.drain()
        except:
            pass
        try:
            writer.close()
        except:
            pass

    def get_run_session(self,label,session_arg,allow_proxied=False):
        if session_arg is None:
            return None
        run_session = stream_load(self.kt_client,session_arg)
        if run_session is None:
            err_level = bcolors.FAIL if not allow_proxied else bcolors.WARNING
            debug(1,label,"This session object has expired and can not be found in the server anymore",session_arg,color=err_level)
            if allow_proxied:
                debug(1,label,"Using KatapultRunSessionProxy as argument",color=bcolors.WARNING)
                try:
                    session_number = int(session_arg['number'])
                    session_id     = session_arg['id'].strip()
                    run_session    = KatapultRunSessionProxy(session_number,session_id)
                    return run_session
                except:
                    debug(1,label,"Could not create proxied session",session_arg,color=bcolors.FAIL)
                    return None
            else:
                return None
        return run_session

    def get_instance(self,label,instance_arg):
        instance = stream_load(self.kt_client,instance_arg)
        if instance is None:
            debug(1,label,"This instance object has expired and can not be found in the server anymore",arg,color=bcolors.FAIL)
            return None
        return instance

    def send_result(self,result):
        print(STREAM_RESULT+json.dumps(stream_dump(result)))

    async def process_command(self,command,args,writer):
        #sock = writer.transport.get_extra_info('socket')
        
        string_writer = ByteStreamWriter(writer)
        sys.stdout = string_writer
        sys.stderr = string_writer

        if self.kt_client is None:
            if command != 'init' and command != 'shutdown':
                print(bcolors.WARNING+"Server not ready. Run 'init CONFIG_FILE' or 'start' command first"+bcolors.ENDC)
                await writer.drain()
                return 
        try:

            if command == 'init':

                if not args:
                    config_ = None
                elif len(args)==1:
                    config_ = args[0].strip() 
                else:
                    config_ = None
                self.kt_client  = kt.get_client(config_)

                await self.wakeup() 

                init_objects = await self.kt_client.get_objects() 
                self.send_result(init_objects)    

            elif command == 'get_objects':

                objs = await self.kt_client.get_objects()
                self.send_result(objs)

            elif command == 'wakeup':

                await self.kt_client.wakeup()

            elif command == 'start':

                if not args:
                    reset = False
                elif len(args)==1:
                    reset = args[0]
                else:
                    reset = False

                await self.kt_client.start(reset)

            elif command == 'cfg_add_instances':
                config = None
                kwargs = dict()
                if args and len(args)>=1:
                    config = args[0]
                elif args and len(args)>=2:
                    config = args[0]
                    kwargs = stream_load(self.kt_client,args[1])
                else:
                    print("Error: you need to send a JSON stream for config")
                    await writer.drain()
                    return
                added_objects = await self.kt_client.cfg_add_instances(config,**kwargs)
                self.send_result(added_objects)

            elif command == 'cfg_add_en((vironments':
                config = None
                kwargs = dict()
                if args and len(args)>=1:
                    config = args[0]
                elif args and len(args)>=2:
                    config = args[0]
                    kwargs = stream_load(self.kt_client,args[1])
                else:
                    print("Error: you need to send a JSON stream for config")
                    await writer.drain()
                    return
                added_objects = await self.kt_client.cfg_add_environments(config,**kwargs)
                self.send_result(added_objects)

            elif command == 'cfg_add_jobs':
                config = None
                kwargs = dict()
                if args and len(args)>=1:
                    config = args[0]
                elif args and len(args)>=2:
                    config = args[0]
                    kwargs = stream_load(self.kt_client,args[1])
                else:
                    print("Error: you need to send a JSON stream for config")
                    await writer.drain()
                    return
                added_objects = await self.kt_client.cfg_add_jobs(config,**kwargs)
                self.send_result(added_objects)

            elif command == 'cfg_add_config':
                config = None
                kwargs = dict()
                if args and len(args)>=1:
                    config = args[0]
                elif args and len(args)>=2:
                    config = args[0]
                    kwargs = stream_load(self.kt_client,args[1])
                else:
                    print("Error: you need to send a JSON stream for config")
                    await writer.drain()
                    return
                added_objects = await self.kt_client.cfg_add_config(config,**kwargs)
                self.send_result(added_objects)

            elif command == 'cfg_reset':

                await self.kt_client.cfg_reset()

            elif command == 'deploy':

                await self.kt_client.deploy()

            elif command == 'run':

                if args and len(args)==1:
                    continue_session = args[0]
                    run_session = await self.kt_client.run(continue_session)
                else:
                    run_session = await self.kt_client.run()

                self.send_result(run_session)
            
            elif command == 'kill':

                if args and len(args)==1:
                    identifier = args[0].strip()
                    await self.kt_client.kill(identifier)
            
            elif command == 'wait':

                job_state   = KatapultProcessState.DONE|KatapultProcessState.ABORTED
                run_session = None
                if args and len(args) >= 1:
                    job_state = int(args[0])
                if args and len(args) >= 2:
                    run_session = self.get_run_session("WAIT:",args[1])

                await self.kt_client.wait(job_state,run_session)

            elif command == 'get_num_active_processes':

                run_session = None
                if args and len(args) >= 1:
                    run_session = self.get_run_session("GET_NUM_ACTIVE_PROCESSES:",args[0])

                result = await self.kt_client.get_num_active_processes(run_session)                
                self.send_result(result)

            elif command == 'get_num_instances':

                result = await self.kt_client.get_num_instances()                
                self.send_result(result)

            elif command == 'get_states' or command == 'get_jobs_states':

                run_session = None
                if args and len(args) == 1:
                    run_session = self.get_run_session("GET STATES:",args[0])
                    last_running_processes = False
                elif args and len(args) == 2:
                    run_session = self.get_run_session("GET STATES:",args[0]) 
                    last_running_processes = args[1]

                result = await self.kt_client.get_jobs_states(run_session,last_running_processes)
                self.send_result(result)

            elif command == 'print_summary' or command == 'print':

                run_session = None
                instance    = None
                if args and len(args) >= 1:
                    run_session = self.get_run_session("PRINT_SUMMARY:",args[0])
                if args and len(args) >= 2:
                    instance = self.get_instance("PRINT_SUMMARY:",args[1])

                debug(2,run_session,instance)

                await self.kt_client.print_jobs_summary(run_session,instance)

            elif command == 'print_aborted' or command == 'print_aborted_logs':

                run_session = None
                instance    = None
                if args and len(args) >= 1:
                    run_session = self.get_run_session("PRINT_SUMMARY:",args[0])
                if args and len(args) >= 2:
                    instance = self.get_instance("PRINT_SUMMARY:",args[1])

                debug(2,run_session,instance)

                await self.kt_client.print_aborted_logs(run_session,instance)

            elif command == 'print_objects':

                await self.kt_client.print_objects()

            elif command == 'clear_results_dir':

                if args and len(args)>0:
                    directory = args[0].strip()
                    await self.kt_client.clear_results_dir(directory)
                else:
                    await self.kt_client.clear_results_dir()

            elif command == 'fetch_results':

                directory   = None
                run_session = None
                use_cached  = True
                use_normal_output = False

                if args and len(args)>=1:
                    directory = args[0].strip()
                    if directory.lower() == 'none':
                        directory = None
                if args and len(args)>=2:
                    run_session = self.get_run_session("FETCH_RESULTS",args[1],True) # allow shallow object return
                if args and len(args)>=3:
                    use_cached = args[2]
                if args and len(args)>=4:
                    use_normal_output = args[3]

                out_dir = await self.kt_client.fetch_results(directory,run_session,use_cached,use_normal_output)
                
                self.send_result(out_dir)

            elif command == 'finalize':
                
                await self.kt_client.finalize()

            elif command == 'shutdown':

                asyncio.get_event_loop().stop()

            elif command == 'test':

                print("TEST")
            
            else:

                print("UNKNOWN COMMAND")
        
        except Exception as e:
            print(e)
            traceback.print_exc()
            await writer.drain()
            #writer.close()
            self.restore_stdio()
            print(e)
            traceback.print_exc()
            raise e

        await writer.drain()
        # to let time for the buffer to be flushed before killing thread and connection
        #time.sleep(1)
        #writer.close()  

    def restore_stdio(self):
        sys.stdout = self.old_stdout
        sys.stderr = self.old_stderr

# run main loop
def main():

    ctxt = ServerContext(sys.argv)

    try:
        asyncio.run( mainloop(ctxt) )
    except RuntimeError as re:
        print(re)

if __name__ == '__main__':
    main()    


import json , os , stat , yaml , shutil , re
from yaml.loader import FullLoader
import subprocess


def run(command_line):
    result = subprocess.run(command_line, shell=True, stdout=subprocess.PIPE)
    output = result.stdout.decode('utf-8')
    return output

bootstraped = True 
errors = []

env_name = os.path.basename(os.getcwd())
#print(env_name)
HOME = os.environ['HOME']

has_mamba = False

mambadir = os.path.join(HOME,'micromamba')
menvsdir = os.path.join(mambadir,'envs')
menvdir  = os.path.join(menvsdir,env_name)

print("\nENV_CHECK",env_name,"\n----------------------------")

if os.path.isfile('environment.yml'):
    history  = os.path.join(menvdir,'conda-meta','history')
    with open(history,'r') as history_file:
        history = history_file.read()
        #print(history)
        with open('environment.yml','r') as yml_file:
            has_mamba = True
            conda_cfg = yaml.load(yml_file, Loader=FullLoader)
            if 'dependencies' in conda_cfg:
                for dependency in conda_cfg.get('dependencies'):
                    if isinstance(dependency,str):
                        dependency = re.sub('[=<>].*', '', dependency)
                        if dependency not in history:
                            dependency = dependency.strip()
                            if not dependency:
                                continue
                            bootstraped = False
                            errors.append("dependency missing in mamba configuration: " + dependency+ " " + env_name)
                    elif isinstance(dependency,dict):
                        for dep in dependency.keys():
                            dep = dep.strip()
                            if not dep:
                                continue
                            dep = re.sub('[=<>].*', '', dep)
                            if dep not in history:
                                bootstraped = False
                                errors.append("dependency missing in mamba configuration ("+env_name+"): " + dep)

if os.path.isfile('requirements.txt'):
    piplist = ""
    pipsrc  = ""
    if has_mamba:
        pip = os.path.join(menvdir,'bin','pip')
        piplist = run(pip + " list")
        #print("piplist mamba",piplist)
        pipsrc = "mamba"
    else:
        venvpip = os.path.join(HOME,'run','.'+env_name,'bin','pip')
        piplist = run(venvpip + " list")
        #print("piplist venv",piplist)
        pipsrc = "venv"
    with open('requirements.txt','r') as txt_file:
        reqs = txt_file.read().split('\n')
        for req in reqs:
            req = req.strip()
            if not req:
                continue
            req = re.sub('\s*[=<>].*', '', req)
            if req not in piplist:
                bootstraped = False
                errors.append("dependency missing in pip configuration ("+pipsrc+" "+env_name+"): " + req)

if os.path.isfile('aptget.sh'):
    with open('aptget.sh','r') as sh_file:
        app_list = sh_file.read().replace('#!/usr/bin/bash\n\nsudo apt-get install -y','')
        app_list = app_list.split(' ')
        for app in app_list:
            app = app.strip()
            if not app:
                continue
            if shutil.which(app) is None:
                bootstraped = False
                errors.append("application not installed with apt-get: " + app + " " + env_name)

if os.path.isfile('env_julia.jl'):
    julia_pkgs = run("julia -e 'using Pkg; println(Pkg.dependencies())'")
    with open('env_julia.jl','r') as julia_file:
        pkg_content = julia_file.read()
        pkg_content = pkg_content.replace("using Pkg\n","")
        pkgs = pkg_content.split('\n')
        for pkg in pkgs:
            pkg = pkg.strip()
            if not pkg:
                continue
            pkg = re.sub(r'Pkg\.add\("([^\)]+)"\)', r'\g<1>', pkg)
            julia_script  = "using Pkg; if \"{0}\" in keys(Pkg.dependencies()) println(\"OK\") else print(\"NOT_OK\") end".format(pkg)
            julia_command = "julia -e '{0}'".format(julia_script)
            # julia_out = run(julia_command)
            # if julia_out != "OK":
            #      bootstraped = False
            #      errors.append("julia package not installed: "+pkg + " " + env_name)
            if pkg not in julia_pkgs:
                bootstraped = False
                errors.append("julia package not installed: "+pkg + " " + env_name)

if bootstraped:
    print('\nALL OK')
    with open('state','w') as file_state:
        file_state.write('bootstraped')
    with open('ready','w') as file_ready:
        file_ready.write("")
else:
    print(env_name+" bootstraping FAILED with errors:")
    for error in errors:
        print(error)
    with open('state','w') as file_state:
        file_state.write('failed')
    with open('errors','w') as file_errors:
        json.dump( errors , file_errors )


from katapult import provider as kt
import asyncio , os , sys
from katapult.core import KatapultProcessState
import traceback
import json

async def tail_loop(script_hash,uid):

    generator = await kt_client.tail(script_hash,uid) 
    for line in generator:
        print(line)


async def mainloop(kt_client,reset=False):

    print("\n== START ==\n")

    # distribute the jobs on the instances (dummy algo for now)
    await kt_client.start(reset) 

    # clear the cache
    await kt_client.clear_results_dir()

    print("\n== DEPLOY ==\n")

    # pre-deploy instance , environments and job files
    # it is recommended to wait here allthough run.sh should wait for bootstraping
    # currently, the bootstraping is non-blocking
    # so this will barely wait ... (the jobs will do the waiting ...)
    await kt_client.deploy()

    print("\n== RUN ==\n")

    # run the scripts and get a process back
    run_session = await kt_client.run()

    #await kt_client.kill( run_session.get_id() )

    print("\n== WAIT ==\n")

    print("Waiting for DONE or ABORTED ...")
    # now that we have 'watch' before 'wait' , this will exit instantaneously
    # because watch includes 'wait' mode intrinsiquely
    await kt_client.wait(KatapultProcessState.DONE|KatapultProcessState.ABORTED,run_session)

    print("\n== SUMMARY ==\n")

    # just to show the API ...
    await kt_client.get_jobs_states()

    # print("\n== WAIT and TAIL ==\n")

    await kt_client.print_aborted_logs()

    print("\n== FETCH RESULTS ==\n")

    await kt_client.fetch_results()

    print("\n== FINALIZE ==\n")

    await kt_client.finalize()

    print("\n== DONE ==\n")

async def cliloop(config_file):

    print("\n== START ==\n")

    # you have to call init before start
    os.system("python3 -m katapult.cli init "+config_file)
    os.system("python3 -m katapult.cli start")

    # clear the cache
    os.system("python3 -m katapult.cli clear_results_dir")

    print("\n== DEPLOY ==\n")

    # pre-deploy instance , environments and job files
    # it is recommended to wait here allthough run.sh should wait for bootstraping
    # currently, the bootstraping is non-blocking
    # so this will barely wait ... (the jobs will do the waiting ...)
    os.system("python3 -m katapult.cli deploy")

    print("\n== RUN ==\n")

    # run the scripts and get a process back
    os.system("python3 -m katapult.cli run")

    #await kt_client.kill( run_session.get_id() )

    print("\n== ADD more ... ==\n")

    if os.path.isfile('config_add.py'):
        print("waiting 30 seconds before adding stuff ...")
        await asyncio.sleep(30)
        print("adding config_add.py")
        os.system("python3 -m katapult.cli cfg_add_config config_add.py")
    else:
        print("this step is optional: you need to have a 'config_add.py' file present")

    print("\n== WAIT ==\n")

    print("Waiting for DONE or ABORTED ...")
    # now that we have 'watch' before 'wait' , this will exit instantaneously
    # because watch includes 'wait' mode intrinsiquely
    os.system("python3 -m katapult.cli wait")

    print("\n== SUMMARY ==\n")

    # just to show the API ...
    os.system("python3 -m katapult.cli get_jobs_states")

    # print("\n== WAIT and TAIL ==\n")

    os.system("python3 -m katapult.cli print_aborted_logs")

    print("\n== FETCH RESULTS ==\n")

    os.system("python3 -m katapult.cli fetch_results")

    print("\n== FINALIZE ==\n")

    os.system("python3 -m katapult.cli finalize")
    # also shutdown the server ...
    os.system("python3 -m katapult.cli shutdown")

    print("\n== DONE ==\n")    

async def waitloop(kt_client):

    print("\n== START ==\n")

    await kt_client.start()

    # clear the cache
    await kt_client.clear_results_dir()

    print("\n== WAIT ==\n")
    
    await kt_client.wait(KatapultProcessState.DONE|KatapultProcessState.ABORTED)

    print("\n== SUMMARY ==\n")

    # just to show the API ...
    await kt_client.get_jobs_states()
    await kt_client.print_aborted_logs()

    print("\n== FETCH RESULTS ==\n")

    await kt_client.fetch_results()

    print("\n== FINALIZE ==\n")

    # we have to wait for the watcher daemon here
    # otherwise the program will exit and the daemon will have a CancelledError 
    await kt_client.finalize()

    print("\n== DONE ==\n")        

# run main loop
def main():

    if len(sys.argv)<=1:
        print("You need to specify a config file path: config.json or config.py")
        sys.exit(1)
    
    config_file = sys.argv[1]
    kt_client   = kt.get_client(config_file)
    command = None
    if len(sys.argv)>2: 
        command = sys.argv[2]
    
    if command == 'wait':
        asyncio.run( waitloop(kt_client) )
    elif command == 'cli':
        asyncio.run( cliloop(config_file) )
    elif command == 'reset':
        asyncio.run( mainloop(kt_client,True) )
    else:
        asyncio.run( mainloop(kt_client) )

if __name__ == '__main__':
    main()    


import asyncio
import shutil
import os
import shlex
import traceback
import fnmatch
from datetime import datetime
from scriptflow.runners import AbstractRunner
from katapult.provider import get_client
from katapult.core import KatapultProcessState
from katapult.utils import generate_unique_id

TASK_PROP_UID = 'katapult_uid'
TASK_DICT_HANDLED = '_handled'
JOB_CFG_T_UID = 'task_uid'
SLEEP_PERIOD_SHORT = 0.1
SLEEP_PERIOD_LONG  = 15

# Runner for tlamadon/scriptflow

class KatapultRunner(AbstractRunner):

    def __init__(self, conf, **kwargs):
        self._katapult         = get_client(conf)
        self._run_session       = None
        self._processes         = {}
        self._num_instances     = 0 
        self._sleep_period      = SLEEP_PERIOD_SHORT
        self._handle_task_queue = kwargs.get('handle_task_queue',True)
        self._do_reset          = kwargs.get('reset',False)
        self._upload_python     = kwargs.get('upload_python_files',False)

    def _move_results(self,results_dir):
        for root, dirs, files in os.walk(results_dir):
            for filename in files:
                moved_file = os.path.join(os.getcwd(),filename)
                if os.path.isfile(moved_file):
                    os.remove(moved_file)
                src_file = os.path.join(results_dir,filename)
                shutil.move(src_file,os.getcwd())
                #print("moved {0} to {1}".format(src_file,moved_file))

            for dirname in dirs:
                moved_dir = os.path.join(os.getcwd(),dirname)
                if os.path.isdir(moved_dir):
                    shutil.rmtree(moved_dir)
                shutil.move(os.path.join(results_dir,dirname),os.getcwd())
        
        # remove it so the next fetch_result doesnt return prematurely (we're using cached=True)
        shutil.rmtree(results_dir)
            

    def _associate_jobs_to_tasks(self,objects):
        # associate the job objects with the tasks
        for (k,p) in self._processes.items():
            if p["job"]:
                continue
            if p.get(TASK_DICT_HANDLED,False)==False:
                continue
            found = False
            for job in objects['jobs']:
                if job.get_config(JOB_CFG_T_UID) == p["task"].get_prop(TASK_PROP_UID):
                    p["job"] = job
                    found = True
                    break
            if not found:
                # this can happen due to asynchronisms
                #raise Error("Internal Error: could not find job for task")        
                pass

    def _flatten_inouts(self,obj,result):
        if isinstance(obj,list):
            for o in obj:
                self._flatten_inouts(o,result)
        elif isinstance(obj,dict):
            for v in obj.values():
                self._flatten_inouts(v,result)
        elif isinstance(obj,str):
            result.append(obj)
        return result 

    def size(self):
        return(len(self._processes))

    def available_slots(self):
        if self._handle_task_queue:
            # always available slots
            # this will cause the controller to add all the tasks at once
            # katapult will handle the stacking with its own batch_run feature ...
            return self._num_instances
        else:
            return self._num_instances - len(self._processes)

    """
        Start tasks
    """

    def add(self, task):

        # we're not doing much here because this method is not async ...
        # leave it to the update(..) method to do the work...
        # we also want to group the handling of adding jobs and running them

        # we don't want to wait as much when we are adding tasks >> lower the period
        self._sleep_period = SLEEP_PERIOD_SHORT 

        # let's not touch the name/uid of the task
        # and it may be empty ...
        task.set_prop(TASK_PROP_UID,generate_unique_id())

        self._processes[task.hash] = {
            "task": task , 
            "job":  None , # we have a one-to-one relationship between job and task (no demultiplier)
            "state" : KatapultProcessState.UNKNOWN ,
            'start_time': datetime.now().strftime("%d/%m/%Y %H:%M:%S")
        }

    """
        Continuously checks on the tasks
    """
    async def loop(self, controller):

        # start the provider and get remote instances information
        await self._katapult.start(self._do_reset)
        # delete the fetch results directory
        await self._katapult.clear_results_dir()

        # cache the number of instances
        self._num_instances = await self._katapult.get_num_instances()

        while True:
            try:
                await self._update(controller)                
                await asyncio.sleep(self._sleep_period)
            except Exception as e:
                print("Loop Error",e)
                traceback.print_exc()
                break

    async def _update(self,controller):

        jobs_cfg = []

        job_added    = True
        while job_added:
            job_added = False
            for (k,p) in self._processes.items():
                # create the job if its not there (it will append/run automatically)
                if p["job"]:
                    continue 

                if p.get(TASK_DICT_HANDLED,False) == True:
                    continue

                task = p["task"]

                nuargs = []
                for arg in task.get_command():
                    if " " in arg:
                        #nuargs.append( "\\\"" + arg + "\\\"")
                        nuargs.append( "\"" + arg + "\"")
                        #nuargs.append(arg)
                    else:
                        nuargs.append(arg)

                job_cfg = {
                    'input_files'  : self._flatten_inouts(task.deps,[]) ,
                    'output_files' : self._flatten_inouts(task.outputs,[]),
                    #'run_command'  : shlex.join(task.get_command()) ,
                    #'run_command'  : " ".join(task.get_command()) ,
                    #'run_command'  : " ".join([shlex.quote(arg) for arg in task.get_command()]) ,
                    'run_command'  : " ".join(nuargs) ,
                    'cpus_req'     : task.ncore ,
                    'number'       : 1 ,
                    JOB_CFG_T_UID  : task.get_prop(TASK_PROP_UID)
                }

                # option to upload all python files of the directory
                # in case some defined routines are used (as in the examples)
                if self._upload_python:
                    filesOfDirectory = os.listdir('.')
                    pattern = "*.py"
                    for file in filesOfDirectory:
                        if fnmatch.fnmatch(file, pattern):
                            if not job_cfg.get('upload_files'):
                                job_cfg['upload_files'] = []
                            job_cfg['upload_files'].append(file)
                # lets upload also the sflow file
                # in case some defined routines are used (as in the examples)
                # if os.path.isfile( os.path.join(os.getcwd(),'sflow.py') ):
                #     if not job_cfg.get('upload_files'):
                #         job_cfg['upload_files'] = []
                #     job_cfg['upload_files'].append('sflow.py')

                jobs_cfg.append( job_cfg )

                job_added     = True
                p[TASK_DICT_HANDLED] = True
            
            # we've just added a task let's see if another one comes in ...
            await asyncio.sleep(.5)

        # add the new jobs and get jobs objects back
        if len(jobs_cfg)>0:

            objects = await self._katapult.cfg_add_jobs(jobs_cfg,run_session=self._run_session)

            # 1 task <-> 1 job
            assert( objects and len(objects['jobs']) == len(jobs_cfg) )

            # run the jobs
            if self._run_session is None:
                await self._katapult.deploy() # deploy the new jobs etc.
                self._run_session = await self._katapult.run() 
            else:
                await self._katapult.deploy() # deploy the new jobs etc.
                await self._katapult.run(True) # continue_session

            # associate task <-> job
            self._associate_jobs_to_tasks(objects)

        if len(self._processes)>0:
            # fetch statusses here ...
            # True stands for 'last_running_processes' meaning we will only get one process per job (the last one)
            processes_states = await self._katapult.get_jobs_states(self._run_session,True)

            # augment the period now ...
            if processes_states and len(processes_states)>0:
                self._sleep_period = SLEEP_PERIOD_LONG

        else:
            self._sleep_period = SLEEP_PERIOD_SHORT


        has_aborted_process = False

        #update status
        to_remove = []
        for (k,p) in self._processes.items():
            job = p["job"]
            if not job: # this task hadnt been handled yet
                continue 
            if p.get(TASK_DICT_HANDLED,False)==False:
                continue
            found_process = False
            for pstatus in processes_states.values():
                if pstatus['job_config'][JOB_CFG_T_UID] == p["task"].get_prop(TASK_PROP_UID):

                    # we match with runner-managed task id, but the job_id should also be right
                    # (we self manage this match in case katapult internal api changes)
                    assert pstatus['job_id'] == job.get_id() , "Internal Error: task<->job<->last_process: it looks like something is wrong ! Fix it"
                    
                    p["state"] = pstatus['state']
                    found_process = True
                    break
            
            # should not happen
            assert found_process , "Internal Error: task<->job<->last_process: We couldn't find a process in the results"

            if p["state"] & ( KatapultProcessState.DONE | KatapultProcessState.ABORTED):
                to_remove.append(k)
                has_aborted_process = has_aborted_process or p["state"] == KatapultProcessState.ABORTED

        # we got some jobs to fetch
        if len(to_remove)>0:
            results_dir = await self._katapult.fetch_results('tmp',self._run_session,True,True)
            self._move_results(results_dir)
            if has_aborted_process:
                await self._katapult.print_aborted_logs(self._run_session)

        # signal completed tasks
        for k in to_remove:
            task = self._processes[k]["task"]
            del self._processes[k]       
            controller.add_completed( task )